==> sqfentity_gen-1.0.2%2B2.tar.gz: lib/sqfentity_gen.dart <==
///****************************   LICENSE   *****************************
/// Copyright (C) 2019, HUSEYIN TOKPUNAR http://huseyintokpinar.com/
/// Download & Update Latest Version: https://github.com/hhtokpinar/sqfEntity
/// Licensed under the Apache License, Version 2.0 (the "License");
/// you may not use this file except in compliance with the License.
/// You may obtain a copy of the License at

///    http://www.apache.org/licenses/LICENSE-2.0

/// Unless required by applicable law or agreed to in writing, software
/// distributed under the License is distributed on an "AS IS" BASIS,
/// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
/// See the License for the specific language governing permissions and
/// limitations under the License.
///**********************************************************************/

import 'package:analyzer/dart/constant/value.dart';

part 'sqfentity_base.dart';

class SqfEntityModelBuilder extends SqfEntityModelBase {
  SqfEntityModelBuilder(this.model, this.instanceName);
  final DartObject model;
  final String instanceName;
//type 'String' is not a subtype of type 'SqfEntitySequenceBase' in type cast

  SqfEntityModelBase toModel() {
    final dbModel = _DbModel()
      ..modelName = getStringValue(model, 'modelName') ?? instanceName
      ..databaseName = getStringValue(model, 'databaseName')
      ..sequences = toSequenceList(getListValue(model, 'sequences'))
      ..databaseTables = toTableList(getListValue(model, 'databaseTables'))
      ..bundledDatabasePath = getStringValue(model, 'bundledDatabasePath');
    return dbModel;
  }

  SqfEntityTableBase toTable(DartObject obj) {
    return SqfEntityTableBase()
      ..tableName = getStringValue(
          obj, 'tableName') // getListValue(obj, 'fields').toString()
      ..primaryKeyName = getStringValue(obj, 'primaryKeyName')
      ..primaryKeyType = getTypeValue(obj, 'primaryKeyType') as PrimaryKeyType
      ..useSoftDeleting = getBoolValue(obj, 'useSoftDeleting')
      ..fields = toFields(getListValue(obj, 'fields'))
      ..defaultJsonUrl = getStringValue(obj, 'defaultJsonUrl')
      ..modelName = getStringValue(obj, 'modelName')
      ..init();
  }

  List<SqfEntityTableBase> toTableList(List<DartObject> objTables) {
    print('SQFENTITY_GEN.DART: recognizing Tables ($databaseName)');
    final sqfEntityTableList = List<SqfEntityTableBase>();
    for (var obj in objTables) {
      sqfEntityTableList.add(toTable(obj));
    }
    return sqfEntityTableList;
  }

  SqfEntitySequenceBase toSequence(DartObject obj) {
    return SqfEntitySequenceBase()
      ..sequenceName = getStringValue(obj, 'sequenceName')
      ..modelName = getStringValue(obj, 'modelName')
      ..init();
  }

  List<SqfEntitySequenceBase> toSequenceList(List<DartObject> objSequences) {
    if (objSequences == null) {
      print('SQFENTITY_GEN.DART: toSequenceList() returned null');
      return null;
    }
    print('SQFENTITY_GEN.DART: recognizing Sequences ($databaseName)');
    final sqfEntitySQEList = List<SqfEntitySequenceBase>();
    for (var obj in objSequences) {
      try {
        sqfEntitySQEList.add(toSequence(obj));
      } catch (e) {
        print(
            'SQFENTITY_GEN.DART: recognizing Sequences ERROR. ($databaseName) tried obj=>${obj.toString()}\nError Message:${e.toString()}');
      }
    }
    return sqfEntitySQEList;
  }

  List<SqfEntityFieldType> toFields(List<DartObject> objFields) {
    final sqfEntityFieldList = List<SqfEntityFieldType>();
    for (var obj in objFields) {
      if (obj.toString().startsWith('SqfEntityFieldRelationship')) {
        sqfEntityFieldList.add(SqfEntityFieldRelationshipBase(
            toTable(obj.getField('parentTable')),
            getTypeValue(obj, 'deleteRule') as DeleteRule)
          ..defaultValue = getDynamicValue(obj, 'defaultValue'));
      } else {
        sqfEntityFieldList.add(SqfEntityFieldBase(
            getStringValue(obj, 'fieldName'),
            getTypeValue(obj, 'dbType') as DbType)
          ..defaultValue = getDynamicValue(obj, 'defaultValue')
          ..sequencedBy = obj
                  .getField('sequencedBy')
                  .toString()
                  .contains('SqfEntitySequenceAnnotation')
              ? toSequence(obj.getField('sequencedBy'))
              : null);
      }
    }
    return sqfEntityFieldList;
  }

  dynamic getDynamicValue(DartObject obj, String name) {
    final String type = obj
        .getField(name)
        .toString()
        .substring(0, obj.getField(name).toString().indexOf(' '))
        .toLowerCase();
    switch (type) {
      case 'string':
        return getStringValue(obj, name);
        break;
      case 'int':
        return getIntValue(obj, name);
        break;
      case 'bool':
        return getBoolValue(obj, name);
        break;
      default:
        return null;
    }
  }

  String getStringValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toStringValue() : null;
  bool getBoolValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toBoolValue() : null;
  int getIntValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toIntValue() : null;
  List<DartObject> getListValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toListValue() : null;
  dynamic getTypeValue(DartObject obj, String name) =>
      ifExist(obj, name) ? convertType(obj.getField(name)) : null;

  bool ifExist(DartObject obj, String name) => obj.getField(name) != null;
}

dynamic convertType(dynamic T) {
  final type = T.type;
  var types = Map<String, dynamic>();
  if (type.toString() == 'DbType') {
    types = dbTypes();
  } else if (type.toString() == 'PrimaryKeyType') {
    types = primaryKeyTypes();
  } else if (type.toString() == 'DeleteRule') {
    types = deleteRuleTypes();
  }
  for (var typ in types.entries) {
    if (T.toString().contains(typ.key)) {
      return typ.value;
    }
  }
  return type.toString();
}

Map<String, dynamic> dbTypes() {
  final types = Map<String, dynamic>();
  for (var typ in DbType.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] = typ;
  }
  return types;
}

Map<String, dynamic> primaryKeyTypes() {
  final types = Map<String, dynamic>();
  for (var typ in PrimaryKeyType.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] = typ;
  }
  return types;
}

Map<String, dynamic> deleteRuleTypes() {
  final types = Map<String, dynamic>();
  for (var typ in DeleteRule.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] = typ;
  }
  return types;
}

class SqfEntityConverter {
  SqfEntityConverter(this._m);
  final SqfEntityModelBase _m;
  String _tableList() {
    if (_m.databaseTables == null || _m.databaseTables.isEmpty) {
      return '';
    }
    final list = StringBuffer()..writeln('databaseTables = [');
    for (final table in _m.databaseTables) {
      list.writeln('Table${table.modelName}.getInstance,');
    }
    list.writeln('];');
    return list.toString();
  }

  String _tableListConst() {
    if (_m.databaseTables == null || _m.databaseTables.isEmpty) {
      return '';
    }
    final list = StringBuffer();
    for (final table in _m.databaseTables) {
      list.writeln('table${table.modelName},');
    }
    return list.toString();
  }

  String _sequenceList() {
    if (_m.sequences == null || _m.sequences.isEmpty) return '';
    final list = StringBuffer()..writeln('sequences = [');

    for (final seq in _m.sequences) {
      list.writeln('Sequence${seq.modelName}.getInstance,');
    }
    list.writeln('];');
    return list.toString();
  }

  String get __createModelTables => _createModelTables();
  String get __createModelTablesConst => _createModelTablesConst();
  String get __createModelSequences => _createModelSequences();
  String get __tableList => _tableList();
  String get __tableListConst => _tableListConst();
  String get __sequenceList => _sequenceList();

  String createModelDatabase() {
    _m.modelName = toModelName(_m.modelName, 'MyDbModel');
    return '''
//  These classes was generated by SqfEntity 
//  Copyright (c) 2019, All rights reserved. Use of this source code is governed by a
//  Apache license that can be found in the LICENSE file.

//  To use these SqfEntity classes do following: 
//  - import model.dart into where to use
//  - start typing ex:${_m.databaseTables.isNotEmpty ? _m.databaseTables[0].modelName : ''}.select()... (add a few filters with fluent methods)...(add orderBy/orderBydesc if you want)...
//  - and then just put end of filters / or end of only select()  toSingle() / or toList() 
//  - you can select one or return List<yourObject> by your filters and orders
//  - also you can batch update or batch delete by using delete/update methods instead of tosingle/tolist methods
//    Enjoy.. Huseyin Tokpunar    

$__createModelTables

$__createModelSequences

// BEGIN DATABASE MODEL
class ${_m.modelName} extends SqfEntityModelProvider {
  ${_m.modelName}() {
    databaseName = '${_m.databaseName}';
    $__tableList
    $__sequenceList
    bundledDatabasePath =
        ${_getNullableStringValue(_m.bundledDatabasePath)}; //'assets/sample.db'; // This value is optional. When bundledDatabasePath is empty then EntityBase creats a new database when initializing the database
  }
}
// END DATABASE MODEL

  ''';
  }

  String createConstDatabase() {
    _m.modelName = toModelName(_m.modelName, 'MyDbModel');
    return '''
//  BEGIN ${_m.databaseName} MODEL
//  Paste the following code into your model.dart
//  Enjoy.. Huseyin Tokpunar    

$__createModelTablesConst

// BEGIN DATABASE MODEL
@SqfEntityBuilder(${tocamelCase(_m.modelName)})
const ${tocamelCase(_m.modelName)} = SqfEntityModel(
    modelName: '${_m.modelName}',
    databaseName: '${_m.databaseName}',
    databaseTables: [$__tableListConst],
    // put defined sequences into the sequences list.
    bundledDatabasePath:
        '${_m.bundledDatabasePath}' 
);
// END ${_m.databaseName} MODEL
''';
  }

  String _createModelTables() {
    if (_m.databaseTables == null) return '';

    final strTables = StringBuffer()..writeln('// BEGIN TABLES');

    for (final table in _m.databaseTables) {
      strTables.writeln('''
// ${table.modelName} TABLE      
class Table${table.modelName} extends SqfEntityTableBase {
  Table${table.modelName}() {
    // declare properties of EntityTable
    tableName = '${table.tableName}';
    primaryKeyName = '${table.primaryKeyName}';
    primaryKeyType = ${table.primaryKeyType.toString()};
    useSoftDeleting = ${table.useSoftDeleting};
    // when useSoftDeleting is true, creates a field named 'isDeleted' on the table, and set to '1' this field when item deleted (does not hard delete)

    // declare fields
    fields = [
      ${_createModelFields(table)}
    ];
    super.init();
  }
  static SqfEntityTableBase _instance;
  static SqfEntityTableBase get getInstance {
    return _instance = _instance ?? Table${table.modelName}();
  }
}''');
    }
    strTables.writeln('// END TABLES');
    return strTables.toString();
  }

  String _createModelTablesConst() {
    if (_m.databaseTables == null || _m.databaseTables.isEmpty) {
      return '';
    }
    final strTables = StringBuffer()..writeln('// BEGIN TABLES');
    for (final table in _m.databaseTables) {
      strTables.writeln('''
const table${toCamelCase(table.tableName)} = SqfEntityTable(
    tableName: '${table.tableName}',
    primaryKeyName: '${table.primaryKeyName}',
    primaryKeyType: ${table.primaryKeyType.toString()},
    useSoftDeleting: false,
    // when useSoftDeleting is true, creates a field named 'isDeleted' on the table, and set to '1' this field when item deleted (does not hard delete)
    fields: [
      ${_createModelFieldsConst(table)}
    ]);''');
    }
    strTables.writeln('// END TABLES');
    return strTables.toString();
  }

  String _createModelSequences() {
    final strSequences = StringBuffer();
    if (_m.sequences == null) return '';
    strSequences.writeln('// BEGIN SEQUENCES');
    for (var seq in _m.sequences) {
      strSequences.writeln('''
// ${seq.sequenceName} SEQUENCE
class Sequence${seq.modelName} extends SqfEntitySequenceBase {
  Sequence${seq.modelName}() {
    sequenceName = '${seq.sequenceName}';
    maxValue = ${seq.maxValue};     /* optional. default is max int (9.223.372.036.854.775.807) */
    cycle = ${seq.cycle};      /* optional. default is false; */
    minValue = ${seq.minValue};    /* optional. default is 0 */
    incrementBy = ${seq.incrementBy}; /* optional. default is 1 */
    startWith = ${seq.startWith};  /* optional. default is 0 */
    super.init();
  }
  static Sequence${seq.modelName} _instance;
  static Sequence${seq.modelName} get getInstance {
    return _instance = _instance ?? Sequence${seq.modelName}();
  }
}''');
    }
    strSequences.writeln('// END SEQUENCES');
    return strSequences.toString();
  }

  String _createModelFields(SqfEntityTableBase table) {
    final strFields = StringBuffer();

    for (final field in table.fields) {
      if (field is SqfEntityFieldRelationshipBase) {
        strFields.writeln(
            'SqfEntityFieldRelationshipBase(Table${field.table.modelName}.getInstance, ${field.deleteRule.toString()}${_getNullableDefaultValue(field.defaultValue)}),');
      } else {
        strFields.writeln(
            'SqfEntityFieldBase(\'${field.fieldName}\', ${field.dbType.toString()}${_getNullableDefaultValue(field.defaultValue)}),');
      }
    }

    return strFields.toString();
  }

  String _createModelFieldsConst(SqfEntityTableBase table) {
    final strFields = StringBuffer();

    for (final field in table.fields) {
      if (field is SqfEntityFieldRelationshipBase) {
        strFields.writeln(
            'SqfEntityFieldRelationship(parentTable: table${field.table.modelName}, deleteRule: DeleteRule.NO_ACTION, fieldName: \'${field.fieldName}\'),');
      } else {
        strFields.writeln(
            'SqfEntityField(\'${field.fieldName}\', ${field.dbType.toString()}),');
      }
    }
    return strFields.toString();
  }

  dynamic _getNullableStringValue(dynamic value) {
    return value is String
        ? '\'${value.toString().replaceAll('\'', '\\\'')}\''
        : value;
  }

  dynamic _getNullableDefaultValue(dynamic value) {
    return value is String
        ? ', defaultValue: \'${value.toString().replaceAll('\'', '\\\'')}\''
        : value != null ? ', defaultValue: $value' : '';
  }

  /// Creates model of tables
  String createEntites() {
    //final String modelName = toString().replaceAll('Instance of', '').replaceAll("'", '').trim();
    final modelString = StringBuffer();
    if (_m.databaseTables != null) {
      modelString.writeln('// BEGIN ENTITIES');
      for (var table in _m.databaseTables) {
        table
          ..dbModel = _m.modelName
          ..collections = _getCollections(table);
        modelString
          ..writeln(SqfEntityObjectBuilder(table).toString())
          ..writeln(SqfEntityObjectField(table).toString())
          ..writeln(SqfEntityObjectFilterBuilder(table).toString())
          ..writeln(SqfEntityFieldBuilder(table).toString())
          ..writeln(SqfEntityObjectManagerBuilder(table).toString());
      }
    }

    if (_m.sequences != null) {
      for (var sequence in _m.sequences) {
        modelString.writeln(SqfEntitySequenceBuilder(sequence).toString());
      }
    }
    modelString..writeln('''class SequenceManager extends SqfEntityProvider {
SequenceManager() : super(${_m.modelName}());
}''')..writeln('// END OF ENTITIES');
    print('SQFENTITIY: [databaseTables] Model was created successfully. ');
    return modelString.toString();
  }

  List<TableCollectionBase> _getCollections(SqfEntityTableBase table) {
    final collectionList = List<TableCollectionBase>();
    for (var _table in _m.databaseTables) {
      for (var field
          in _table.fields.where((f) => f is SqfEntityFieldRelationshipBase)) {
        if (field.table.tableName == table.tableName) {
          collectionList.add(TableCollectionBase(
              _table, field as SqfEntityFieldRelationshipBase));
        }
      }
    }
    return collectionList;
  }
}

class _DbModel extends SqfEntityModelBase {}

class MyStringBuffer extends StringBuffer {
  void printToDebug(String text) {
    print(text);
  }
}

==> sqfentity_gen-1.0.3%2B2.tar.gz: lib/sqfentity_gen.dart <==
///****************************   LICENSE   *****************************
/// Copyright (C) 2019, HUSEYIN TOKPUNAR http://huseyintokpinar.com/
/// Download & Update Latest Version: https://github.com/hhtokpinar/sqfEntity
/// Licensed under the Apache License, Version 2.0 (the "License");
/// you may not use this file except in compliance with the License.
/// You may obtain a copy of the License at

///    http://www.apache.org/licenses/LICENSE-2.0

/// Unless required by applicable law or agreed to in writing, software
/// distributed under the License is distributed on an "AS IS" BASIS,
/// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
/// See the License for the specific language governing permissions and
/// limitations under the License.
///**********************************************************************/

import 'package:analyzer/dart/constant/value.dart';
import 'package:sqfentity_base/sqfentity_base.dart';

//part 'sqfentity_base.dart';

class SqfEntityModelBuilder extends SqfEntityModelBase {
  SqfEntityModelBuilder(this.model, this.instanceName);
  final DartObject model;
  final String instanceName;
//type 'String' is not a subtype of type 'SqfEntitySequenceBase' in type cast

  SqfEntityModelBase toModel() {
    final dbModel = _DbModel()
      ..modelName = getStringValue(model, 'modelName') ?? instanceName
      ..databaseName = getStringValue(model, 'databaseName')
      ..sequences = toSequenceList(getListValue(model, 'sequences'))
      ..databaseTables = toTableList(getListValue(model, 'databaseTables'))
      ..bundledDatabasePath = getStringValue(model, 'bundledDatabasePath');
    return dbModel;
  }

  SqfEntityTableBase toTable(DartObject obj) {
    return SqfEntityTableBase()
      ..tableName = getStringValue(
          obj, 'tableName') // getListValue(obj, 'fields').toString()
      ..primaryKeyName = getStringValue(obj, 'primaryKeyName')
      ..primaryKeyType = getTypeValue(obj, 'primaryKeyType') as PrimaryKeyType
      ..useSoftDeleting = getBoolValue(obj, 'useSoftDeleting')
      ..fields = toFields(getListValue(obj, 'fields'))
      ..defaultJsonUrl = getStringValue(obj, 'defaultJsonUrl')
      ..modelName = getStringValue(obj, 'modelName')
      ..init();
  }

  List<SqfEntityTableBase> toTableList(List<DartObject> objTables) {
    print('SQFENTITY_GEN.DART: recognizing Tables ($modelName)');
    final sqfEntityTableList = List<SqfEntityTableBase>();
    for (var obj in objTables) {
      sqfEntityTableList.add(toTable(obj));
    }
    return sqfEntityTableList;
  }

  SqfEntitySequenceBase toSequence(DartObject obj) {
    return SqfEntitySequenceBase()
      ..sequenceName = getStringValue(obj, 'sequenceName')
      ..modelName = getStringValue(obj, 'modelName')
      ..init();
  }

  List<SqfEntitySequenceBase> toSequenceList(List<DartObject> objSequences) {
    if (objSequences == null) {
      print('SQFENTITY_GEN.DART: toSequenceList() returned null');
      return null;
    }
    print('SQFENTITY_GEN.DART: recognizing Sequences ($modelName)');
    final sqfEntitySQEList = List<SqfEntitySequenceBase>();
    for (var obj in objSequences) {
      try {
        sqfEntitySQEList.add(toSequence(obj));
      } catch (e) {
        print(
            'SQFENTITY_GEN.DART: recognizing Sequences ERROR. ($modelName) tried obj=>${obj.toString()}\nError Message:${e.toString()}');
      }
    }
    return sqfEntitySQEList;
  }

  List<SqfEntityFieldType> toFields(List<DartObject> objFields) {
    final sqfEntityFieldList = List<SqfEntityFieldType>();
    for (var obj in objFields) {
      if (obj.toString().startsWith('SqfEntityFieldRelationship')) {
        sqfEntityFieldList.add(SqfEntityFieldRelationshipBase(
            toTable(obj.getField('parentTable')),
            getTypeValue(obj, 'deleteRule') as DeleteRule)
          ..defaultValue = getDynamicValue(obj, 'defaultValue'));
      } else {
        sqfEntityFieldList.add(SqfEntityFieldBase(
            getStringValue(obj, 'fieldName'),
            getTypeValue(obj, 'dbType') as DbType)
          ..defaultValue = getDynamicValue(obj, 'defaultValue')
          ..sequencedBy = obj
                  .getField('sequencedBy')
                  .toString()
                  .contains('SqfEntitySequenceAnnotation')
              ? toSequence(obj.getField('sequencedBy'))
              : null);
      }
    }
    return sqfEntityFieldList;
  }

  dynamic getDynamicValue(DartObject obj, String name) {
    final String type = obj
        .getField(name)
        .toString()
        .substring(0, obj.getField(name).toString().indexOf(' '))
        .toLowerCase();
    switch (type) {
      case 'string':
        return getStringValue(obj, name);
        break;
      case 'int':
        return getIntValue(obj, name);
        break;
      case 'bool':
        return getBoolValue(obj, name);
        break;
      default:
        return null;
    }
  }

  String getStringValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toStringValue() : null;
  bool getBoolValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toBoolValue() : null;
  int getIntValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toIntValue() : null;
  List<DartObject> getListValue(DartObject obj, String name) =>
      ifExist(obj, name) ? obj.getField(name).toListValue() : null;
  dynamic getTypeValue(DartObject obj, String name) =>
      ifExist(obj, name) ? convertType(obj.getField(name)) : null;

  bool ifExist(DartObject obj, String name) => obj.getField(name) != null;
}

dynamic convertType(dynamic T) {
  final type = T.type;
  var types = Map<String, dynamic>();
  if (type.toString() == 'DbType') {
    types = dbTypes();
  } else if (type.toString() == 'PrimaryKeyType') {
    types = primaryKeyTypes();
  } else if (type.toString() == 'DeleteRule') {
    types = deleteRuleTypes();
  }
  for (var typ in types.entries) {
    if (T.toString().contains(typ.key)) {
      return typ.value;
    }
  }
  return type.toString();
}

Map<String, dynamic> dbTypes() {
  final types = Map<String, dynamic>();
  for (var typ in DbType.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] = typ;
  }
  return types;
}

Map<String, dynamic> primaryKeyTypes() {
  final types = Map<String, dynamic>();
  for (var typ in PrimaryKeyType.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] = typ;
  }
  return types;
}

Map<String, dynamic> deleteRuleTypes() {
  final types = Map<String, dynamic>();
  for (var typ in DeleteRule.values) {
    types[typ.toString().substring(typ.toString().indexOf('.') + 1)] = typ;
  }
  return types;
}

class SqfEntityConverter {
  SqfEntityConverter(this._m);
  final SqfEntityModelBase _m;
  String _tableList() {
    if (_m.databaseTables == null || _m.databaseTables.isEmpty) {
      return '';
    }
    final list = StringBuffer()..writeln('databaseTables = [');
    for (final table in _m.databaseTables) {
      list.writeln('Table${table.modelName}.getInstance,');
    }
    list.writeln('];');
    return list.toString();
  }

  String _tableListConst() {
    if (_m.databaseTables == null || _m.databaseTables.isEmpty) {
      return '';
    }
    final list = StringBuffer();
    for (final table in _m.databaseTables) {
      list.writeln('table${table.modelName},');
    }
    return list.toString();
  }

  String _sequenceList() {
    if (_m.sequences == null || _m.sequences.isEmpty) return '';
    final list = StringBuffer()..writeln('sequences = [');

    for (final seq in _m.sequences) {
      list.writeln('Sequence${seq.modelName}.getInstance,');
    }
    list.writeln('];');
    return list.toString();
  }

  String get __createModelTables => _createModelTables();
  String get __createModelTablesConst => _createModelTablesConst();
  String get __createModelSequences => _createModelSequences();
  String get __tableList => _tableList();
  String get __tableListConst => _tableListConst();
  String get __sequenceList => _sequenceList();

  String createModelDatabase() {
    _m.modelName = toModelName(_m.modelName, 'MyDbModel');
    return '''
//  These classes was generated by SqfEntity 
//  Copyright (c) 2019, All rights reserved. Use of this source code is governed by a
//  Apache license that can be found in the LICENSE file.

//  To use these SqfEntity classes do following: 
//  - import model.dart into where to use
//  - start typing ex:${_m.databaseTables.isNotEmpty ? _m.databaseTables[0].modelName : ''}.select()... (add a few filters with fluent methods)...(add orderBy/orderBydesc if you want)...
//  - and then just put end of filters / or end of only select()  toSingle() / or toList() 
//  - you can select one or return List<yourObject> by your filters and orders
//  - also you can batch update or batch delete by using delete/update methods instead of tosingle/tolist methods
//    Enjoy.. Huseyin Tokpunar    

$__createModelTables

$__createModelSequences

// BEGIN DATABASE MODEL
class ${_m.modelName} extends SqfEntityModelProvider {
  ${_m.modelName}() {
    databaseName = '${_m.databaseName}';
    $__tableList
    $__sequenceList
    bundledDatabasePath =
        ${_getNullableStringValue(_m.bundledDatabasePath)}; //'assets/sample.db'; // This value is optional. When bundledDatabasePath is empty then EntityBase creats a new database when initializing the database
  }
}
// END DATABASE MODEL

  ''';
  }

  String createConstDatabase() {
    _m.modelName = toModelName(_m.modelName, 'MyDbModel');
    return '''
//  BEGIN ${_m.databaseName} MODEL
//  Paste the following code into your model.dart
//  Enjoy.. Huseyin Tokpunar    

$__createModelTablesConst

// BEGIN DATABASE MODEL
@SqfEntityBuilder(${tocamelCase(_m.modelName)})
const ${tocamelCase(_m.modelName)} = SqfEntityModel(
    modelName: '${_m.modelName}',
    databaseName: '${_m.databaseName}',
    databaseTables: [$__tableListConst],
    // put defined sequences into the sequences list.
    bundledDatabasePath:
        '${_m.bundledDatabasePath}' 
);
// END ${_m.databaseName} MODEL
''';
  }

  String _createModelTables() {
    if (_m.databaseTables == null) return '';

    final strTables = StringBuffer()..writeln('// BEGIN TABLES');

    for (final table in _m.databaseTables) {
      strTables.writeln('''
// ${table.modelName} TABLE      
class Table${table.modelName} extends SqfEntityTableBase {
  Table${table.modelName}() {
    // declare properties of EntityTable
    tableName = '${table.tableName}';
    primaryKeyName = '${table.primaryKeyName}';
    primaryKeyType = ${table.primaryKeyType.toString()};
    useSoftDeleting = ${table.useSoftDeleting};
    // when useSoftDeleting is true, creates a field named 'isDeleted' on the table, and set to '1' this field when item deleted (does not hard delete)

    // declare fields
    fields = [
      ${_createModelFields(table)}
    ];
    super.init();
  }
  static SqfEntityTableBase _instance;
  static SqfEntityTableBase get getInstance {
    return _instance = _instance ?? Table${table.modelName}();
  }
}''');
    }
    strTables.writeln('// END TABLES');
    return strTables.toString();
  }

  String _createModelTablesConst() {
    if (_m.databaseTables == null || _m.databaseTables.isEmpty) {
      return '';
    }
    final strTables = StringBuffer()..writeln('// BEGIN TABLES');
    for (final table in _m.databaseTables) {
      strTables.writeln('''
const table${toCamelCase(table.tableName)} = SqfEntityTable(
    tableName: '${table.tableName}',
    primaryKeyName: '${table.primaryKeyName}',
    primaryKeyType: ${table.primaryKeyType.toString()},
    useSoftDeleting: false,
    // when useSoftDeleting is true, creates a field named 'isDeleted' on the table, and set to '1' this field when item deleted (does not hard delete)
    fields: [
      ${_createModelFieldsConst(table)}
    ]);''');
    }
    strTables.writeln('// END TABLES');
    return strTables.toString();
  }

  String _createModelSequences() {
    final strSequences = StringBuffer();
    if (_m.sequences == null) return '';
    strSequences.writeln('// BEGIN SEQUENCES');
    for (var seq in _m.sequences) {
      strSequences.writeln('''
// ${seq.sequenceName} SEQUENCE
class Sequence${seq.modelName} extends SqfEntitySequenceBase {
  Sequence${seq.modelName}() {
    sequenceName = '${seq.sequenceName}';
    maxValue = ${seq.maxValue};     /* optional. default is max int (9.223.372.036.854.775.807) */
    cycle = ${seq.cycle};      /* optional. default is false; */
    minValue = ${seq.minValue};    /* optional. default is 0 */
    incrementBy = ${seq.incrementBy}; /* optional. default is 1 */
    startWith = ${seq.startWith};  /* optional. default is 0 */
    super.init();
  }
  static Sequence${seq.modelName} _instance;
  static Sequence${seq.modelName} get getInstance {
    return _instance = _instance ?? Sequence${seq.modelName}();
  }
}''');
    }
    strSequences.writeln('// END SEQUENCES');
    return strSequences.toString();
  }

  String _createModelFields(SqfEntityTableBase table) {
    final strFields = StringBuffer();

    for (final field in table.fields) {
      if (field is SqfEntityFieldRelationshipBase) {
        strFields.writeln(
            'SqfEntityFieldRelationshipBase(Table${field.table.modelName}.getInstance, ${field.deleteRule.toString()}${_getNullableDefaultValue(field.defaultValue)}),');
      } else {
        strFields.writeln(
            'SqfEntityFieldBase(\'${field.fieldName}\', ${field.dbType.toString()}${_getNullableDefaultValue(field.defaultValue)}),');
      }
    }

    return strFields.toString();
  }

  String _createModelFieldsConst(SqfEntityTableBase table) {
    final strFields = StringBuffer();

    for (final field in table.fields) {
      if (field is SqfEntityFieldRelationshipBase) {
        strFields.writeln(
            'SqfEntityFieldRelationship(parentTable: table${field.table.modelName}, deleteRule: DeleteRule.NO_ACTION, fieldName: \'${field.fieldName}\'),');
      } else {
        strFields.writeln(
            'SqfEntityField(\'${field.fieldName}\', ${field.dbType.toString()}),');
      }
    }
    return strFields.toString();
  }

  dynamic _getNullableStringValue(dynamic value) {
    return value is String
        ? '\'${value.toString().replaceAll('\'', '\\\'')}\''
        : value;
  }

  dynamic _getNullableDefaultValue(dynamic value) {
    return value is String
        ? ', defaultValue: \'${value.toString().replaceAll('\'', '\\\'')}\''
        : value != null ? ', defaultValue: $value' : '';
  }

  /// Creates model of tables
  String createEntites() {
    //final String modelName = toString().replaceAll('Instance of', '').replaceAll("'", '').trim();
    final modelString = StringBuffer();
    if (_m.databaseTables != null) {
      modelString.writeln('// BEGIN ENTITIES');
      for (var table in _m.databaseTables) {
        table
          ..dbModel = _m.modelName
          ..collections = _getCollections(table);
        modelString
          ..writeln(SqfEntityObjectBuilder(table).toString())
          ..writeln(SqfEntityObjectField(table).toString())
          ..writeln(SqfEntityObjectFilterBuilder(table).toString())
          ..writeln(SqfEntityFieldBuilder(table).toString())
          ..writeln(SqfEntityObjectManagerBuilder(table).toString());
      }
    }

    if (_m.sequences != null) {
      for (var sequence in _m.sequences) {
        modelString.writeln(SqfEntitySequenceBuilder(sequence).toString());
      }
    }
    modelString..writeln('''class SequenceManager extends SqfEntityProvider {
SequenceManager() : super(${_m.modelName}());
}''')..writeln('// END OF ENTITIES');
    print('SQFENTITIY: [databaseTables] Model was created successfully. ');
    return modelString.toString();
  }

  List<TableCollectionBase> _getCollections(SqfEntityTableBase table) {
    final collectionList = List<TableCollectionBase>();
    for (var _table in _m.databaseTables) {
      for (var field
          in _table.fields.where((f) => f is SqfEntityFieldRelationshipBase)) {
        if (field.table.tableName == table.tableName) {
          collectionList.add(TableCollectionBase(
              _table, field as SqfEntityFieldRelationshipBase));
        }
      }
    }
    return collectionList;
  }
}

class _DbModel extends SqfEntityModelBase {}

class MyStringBuffer extends StringBuffer {
  void printToDebug(String text) {
    print(text);
  }
}
